# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import subprocess
try:
    # For Python 3.0 and later
    from urllib.request import urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen


from .cached_objects import CachedObject
from .errors import PeeringDBError, PeeringDBNoInfoError


class PeeringDBInfo(CachedObject):

    def _get_peeringdb_url(self):
        raise NotImplementedError()

    @staticmethod
    def _read_from_url(url):
        try:
            response = urlopen(url)
        except Exception as e:
            raise PeeringDBError(
                "Error while retrieving info from PeeringDB: {}".format(
                    str(e)
                )
            )

        return response.read().decode("utf-8")

    def _get_data_from_peeringdb(self):
        plain_text = self._read_from_url(self._get_peeringdb_url())
        try:
            data = json.loads(plain_text)
            return data
        except Exception as e:
            raise PeeringDBError(
                "Error while decoding PeeringDB output: {}".format(
                    str(e)
                )
            )

    def _get_data(self):
        data = self._get_data_from_peeringdb()
        if not "data" in data:
            raise PeeringDBNoInfoError("Missing 'data'")
        if not isinstance(data["data"], list):
            raise PeeringDBNoInfoError("Unexpected format: 'data' is not a list")
        if len(data["data"]) == 0:
            raise PeeringDBNoInfoError("No data for this nextwork")

        return data["data"]

class PeeringDBNet(PeeringDBInfo):

    PEERINGDB_URL = "https://www.peeringdb.com/api/net?asn={asn}"

    def __init__(self, asn, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)
        self.asn = asn

        logging.debug("Getting data from PeeringDB: net {}".format(self.asn))

        self.load_data()
    
        self.info_prefixes4 = self.raw_data[0].get("info_prefixes4", None)
        self.info_prefixes6 = self.raw_data[0].get("info_prefixes6", None)
        self.irr_as_set = self.raw_data[0].get("irr_as_set", None)

    def _get_object_filename(self):
        return "peeringdb_net_{}.json".format(self.asn)

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format(asn=self.asn)

class PeeringDBNetIXLan(PeeringDBInfo):

    PEERINGDB_URL = "https://www.peeringdb.com/api/netixlan?ixlan_id={ixlanid}"

    def __init__(self, ixlanid, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)
        self.ixlanid = ixlanid

        logging.debug("Getting data from PeeringDB: Net IX LAN {}".format(self.ixlanid))

        self.load_data()

    def _get_object_filename(self):
        return "peeringdb_ixlanid_{}.json".format(self.ixlanid)

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format(ixlanid=self.ixlanid)


def clients_from_peeringdb(netixlanid, cache_dir):
    clients = []

    netixlans = PeeringDBNetIXLan(netixlanid, cache_dir=cache_dir).raw_data
    for netixlan in netixlans:
        if netixlan["is_rs_peer"] is True:
            client = {
                "asn": netixlan["asn"],
                "ip": [],
            }
            for ipver in ("ipaddr4", "ipaddr6"):
                if netixlan[ipver]:
                    client["ip"].append(netixlan[ipver].encode("ascii", "ignore"))
            clients.append(client)

    asns = {}

    for client in clients:
        asn = client["asn"]
        net = PeeringDBNet(asn)

        irr_as_sets = net.irr_as_set
        if not irr_as_sets:
            continue

        if "/" in irr_as_sets:
            irr_as_sets = irr_as_sets.split("/")
        else:
            irr_as_sets = [irr_as_sets]

        key = "AS{}".format(asn)
        if key not in asns:
            asns[key] = {
                "as_sets": []
            }

        for irr_as_set in irr_as_sets:
            irr_as_set = irr_as_set.strip()
            if irr_as_set not in asns[key]["as_sets"]:
                asns[key]["as_sets"].append(irr_as_set.encode("ascii", "ignore"))

    data = {
        "asns": asns,
        "clients": clients
    }

    return data



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance

class RichConfigExampleScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    AS_SET = {
        "AS3333": [3333],
        "AS10745": [10745],
    }
    R_SET = {
        "AS10745": [
            "AS10745_allowed_prefixes"
        ],
        "AS3333": [
            "AS3333_allowed_prefixes"
        ],
    }

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance()
        ]

    def set_instance_variables(self):
        self.rs = self._get_instance_by_name("rs")
        
    def test_010_setup(self):
        """{}: instances setup"""
        pass

class RichConfigExampleScenarioBIRD(RichConfigExampleScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     local_files=["client"]),
                    "/etc/bird/bird.conf"
                ),
                (
                    cls.use_static_file("bird.client.local"),
                    "/etc/bird/client.local"
                )
            ]
        )

class RichConfigExampleScenarioOpenBGPD(LiveScenario_TagRejectPolicy,
                                        RichConfigExampleScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None),
                    "/etc/bgpd.conf"
                )
            ]
        )



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from Queue import Queue, Empty, Full
import threading

from ..errors import BuilderError, ARouteServerError

class BaseConfigEnricherThread(threading.Thread):

    DESCR = None

    def __init__(self, tasks_q, errors_q, lock):
        threading.Thread.__init__(self)

        self.tasks_q = tasks_q
        self.errors_q = errors_q
        self.lock = lock

    def do_task(self, task):
        raise NotImplementedError()

    def save_data(self, task, data):
        raise NotImplementedError()

    def run(self):
        logging.debug("{} thread {} started".format(self.DESCR, self.name))

        while True:
            try:
                task = self.tasks_q.get(block=True, timeout=0.1)
            except Empty:
                break

            try:
                data = self.do_task(task)
                if data:
                    with self.lock:
                        self.save_data(task, data)
            except Exception as e:
                if not isinstance(e, ARouteServerError):
                    logging.debug("{} thread {} error: {}".format(
                        self.DESCR, self.name, str(e)))
                try:
                    self.errors_q.put_nowait(True)
                except Full:
                    pass

            self.tasks_q.task_done()

        logging.debug("{} thread {} stopped".format(
            self.DESCR, self.name))

class BaseConfigEnricher(object):

    WORKER_THREAD_CLASS = None

    def __init__(self, builder, threads):
        self.builder = builder
        self.threads = threads
        self.tasks_q = Queue()
        self.errors_q = Queue(maxsize=1)

    def prepare(self):
        pass

    def _config_thread(self, thread):
        pass

    def add_tasks(self):
        raise NotImplementedError()

    def enrich(self):
        self.prepare()

        lock = threading.Lock()

        threads = []
        for i in range(self.threads):
            t = self.WORKER_THREAD_CLASS(
                self.tasks_q, self.errors_q, lock
            )
            self._config_thread(t)
            threads.append(t)

        self.add_tasks()

        for t in threads:
            t.start()

        self.tasks_q.join()

        try:
            self.errors_q.get_nowait()
            raise BuilderError()
        except Empty:
            return



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPD60Instance
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance

class TagASSetScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            cls.CLIENT_INSTANCE_CLASS(
                "AS1",
                cls.DATA["AS1_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS4",
                cls.DATA["AS4_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS4.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS5",
                cls.DATA["AS5_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS5.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
        ]

    def set_instance_variables(self):
        self.rs = self._get_instance_by_name("rs")
        self.AS1 = self._get_instance_by_name("AS1")
        self.AS2 = self._get_instance_by_name("AS2")
        self.AS4 = self._get_instance_by_name("AS4")
        self.AS5 = self._get_instance_by_name("AS5")
        
    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS4)
        self.session_is_up(self.rs, self.AS5)

class TagASSetScenario_WithAS_SETs(object):

    AS_SET = {
        "AS1": [1],
        "AS-AS2": [2],
        "AS-AS4": [4],
        "AS-AS5": [5],
    }
    R_SET = {
        "AS1": [
            "AS1_allowed_prefixes"
        ],
        "AS-AS2": [
            "AS2_allowed_prefixes"
        ],
        "AS-AS4": [
            "AS4_allowed_prefixes"
        ],
        "AS-AS5": [
            "AS5_allowed_prefixes"
        ]
    }

    def _set_lrg_comms(self, lst):
        if isinstance(self.rs, OpenBGPD60Instance):
            return []
        return lst

    def test_030_AS2_prefix_ok_origin_ok(self):
        """{}: AS2 prefix ok origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ok_origin_ok1"], self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514"],
                           lrg_comms=lrg_comms)

    def test_030_AS2_prefix_ko_origin_ok(self):
        """{}: AS2 prefix ko origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ko_origin_ok1"], self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514"],
                           lrg_comms=lrg_comms)

    def test_030_AS2_prefix_ok_origin_ko(self):
        """{}: AS2 prefix ok origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS3_pref_ok_origin_ko1"], self.AS2, as_path="2 3", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64515"],
                           lrg_comms=lrg_comms)

    def test_030_AS2_prefix_ko_origin_ko(self):
        """{}: AS2 prefix ko origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS3_pref_ko_origin_ko1"], self.AS2, as_path="2 3", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           lrg_comms=lrg_comms)

    def test_040_AS4_prefix_ok_origin_ok(self):
        """{}: AS4 prefix ok origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ok_origin_ok1"], self.AS4, as_path="4", next_hop=self.AS4,
                           std_comms=["999:64512", "999:64514"],
                           lrg_comms=lrg_comms)

    def test_040_AS4_prefix_ko_origin_ok(self):
        """{}: AS4 prefix ko origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ko_origin_ok1"], self.AS4, as_path="4", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64514"],
                           lrg_comms=lrg_comms)

    def test_040_AS4_origin_filtered(self):
        """{}: AS4 route filtered (origin ko)"""
        self.receive_route(self.rs, self.DATA["AS3_pref_ok_origin_ko2"],
                           self.AS4, as_path="4 3", next_hop=self.AS4,
                           filtered=True, reject_reason=9)

    def test_040_AS4_prefix_origin_filtered(self):
        """{}: AS4 route filtered (prefix ko, origin ko)"""
        self.receive_route(self.rs, self.DATA["AS3_pref_ko_origin_ko1"],
                           self.AS4, as_path="4 3", next_hop=self.AS4,
                           filtered=True, reject_reason=9)

    def test_050_AS5_prefix_ok_origin_ok(self):
        """{}: AS5 prefix ok origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS5_pref_ok_origin_ok1"], self.AS5, as_path="5", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64514"],
                           lrg_comms=lrg_comms)

    def test_050_AS5_prefix_filtered(self):
        """{}: AS5 route filtered (prefix ko)"""
        self.receive_route(self.rs, self.DATA["AS5_pref_ko_origin_ok1"],
                           self.AS5, as_path="5", next_hop=self.AS5,
                           filtered=True, reject_reason=12)

    def test_050_AS5_prefix_ok_origin_ko(self):
        """{}: AS5 prefix ok origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS3_pref_ok_origin_ko3"], self.AS5, as_path="5 3", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64515"],
                           lrg_comms=lrg_comms)

    def test_050_AS5_origin_filtered(self):
        """{}: AS5 route filtered (prefix ko, origin ko)"""
        # AS5 is configured with prefix enforcement only, that's why
        # the reject reason should be 12 (prefix not in IRRDBs) and
        # not 9 (origin ASN not in IRRDBs).
        self.receive_route(self.rs, self.DATA["AS3_pref_ko_origin_ko1"],
                           self.AS5, as_path="5 3", next_hop=self.AS5,
                           filtered=True, reject_reason=12)

class TagASSetScenario_EmptyAS_SETs(object):

    AS_SET = {
        "AS1": [],
        "AS-AS2": [],
        "AS-AS4": [],
        "AS-AS5": [],
    }
    R_SET = {
        "AS1": [
        ],
        "AS-AS2": [
        ],
        "AS-AS4": [
        ],
        "AS-AS5": [
        ]
    }

    def _set_lrg_comms(self, lst):
        if isinstance(self.rs, OpenBGPD60Instance):
            return []
        return lst

    def test_030_AS2_no_enforcement(self):
        """{}: AS2 no enforcement, prefix and origin not in AS-SET"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        for pref in (self.DATA["AS2_pref_ok_origin_ok1"],
                     self.DATA["AS2_pref_ko_origin_ok1"],
                     self.DATA["AS3_pref_ok_origin_ko1"],
                     self.DATA["AS3_pref_ko_origin_ko1"]):
            self.receive_route(self.rs, pref, self.AS2, next_hop=self.AS2,
                            std_comms=["999:64513", "999:64515"],
                            lrg_comms=lrg_comms)

    def test_040_AS4_origin_enforcement(self):
        """{}: AS4 origin enforcement"""
        for pref in (self.DATA["AS4_pref_ok_origin_ok1"],
                     self.DATA["AS4_pref_ko_origin_ok1"],
                     self.DATA["AS3_pref_ok_origin_ko2"],
                     self.DATA["AS3_pref_ko_origin_ko1"]):
            self.receive_route(self.rs, pref, self.AS4, next_hop=self.AS4,
                               filtered=True, reject_reason=(9, 12))

    def test_050_AS4_prefix_enforcement(self):
        """{}: AS4 prefix enforcement"""
        for pref in (self.DATA["AS5_pref_ok_origin_ok1"],
                     self.DATA["AS5_pref_ko_origin_ok1"],
                     self.DATA["AS3_pref_ok_origin_ko3"],
                     self.DATA["AS3_pref_ko_origin_ko1"]):
            self.receive_route(self.rs, pref, self.AS5, next_hop=self.AS5,
                               filtered=True, reject_reason=(9, 12))

class TagASSetScenarioBIRD(TagASSetScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER),
                    "/etc/bird/bird.conf"
                )
            ]
        )

class TagASSetScenarioOpenBGPD(LiveScenario_TagRejectPolicy, TagASSetScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                )
            ]
        )

class TagASSetScenarioOpenBGPD60(TagASSetScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = "6.0"

class TagASSetScenarioOpenBGPD61(TagASSetScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = "6.1"



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from base import PathHidingScenario_MitigationOn, \
                 PathHidingScenario_MitigationOff, \
                 PathHidingScenarioBIRD
from data4 import PathHidingScenario_Data4
from pierky.arouteserver.tests.live_tests.bird import BIRDInstanceIPv4

class PathHidingScenario_MitigationOn_BIRDIPv4(PathHidingScenario_Data4,
                                               PathHidingScenario_MitigationOn,
                                               PathHidingScenarioBIRD):
    __test__ = True

    RS_INSTANCE_CLASS = BIRDInstanceIPv4
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv4
    IP_VER = 4

    SHORT_DESCR = "Live test, BIRD, path hiding, mitigation on, IPv4"

class PathHidingScenario_MitigationOff_BIRDIPv4(PathHidingScenario_Data4,
                                                PathHidingScenario_MitigationOff,
                                                PathHidingScenarioBIRD):
    __test__ = True

    RS_INSTANCE_CLASS = BIRDInstanceIPv4
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv4
    IP_VER = 4

    SHORT_DESCR = "Live test, BIRD, path hiding, mitigation off, IPv4"




# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

def ask(text, options=None, default=None):
    """Returns: ([True|False], answer)"""

    msg = "{} ".format(text)
    if options:
        msg_options = []
        for opt in options:
            if opt == default:
                msg_options.append(opt.upper())
            else:
                msg_options.append(str(opt))
        msg += "["
        msg += "/".join(msg_options)
        msg += "] "
    else:
        if default:
            msg += "(default: {}) ".format(default)
    sys.stdout.write(msg)

    try:
        answer = raw_input()
    except:
        return False, None

    answer = answer.strip()
    if answer:
        if options and answer not in options:
            print("Invalid choice: {} - must be one of {}.".format(
                answer, ", ".join(options)))
            return False, None
        return True, answer
    else:
        if default:
            return True, default
        else:
            print("No answer given.")
            return False, None

def ask_yes_no(text, default=None):
    return ask(text, ["yes", "no"], default)

def ask_int(text, default=None):
    answer_given, v = ask(text, None, default)
    if not answer_given:
        return False, None
    if not v.isdigit():
        print("Invalid input: it must be an integer.")
        return False, None
    return True, int(v)



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

import yaml

from .cfg_base import TestConfigParserBase 
from pierky.arouteserver.config.clients import ConfigParserClients
from pierky.arouteserver.config.general import ConfigParserGeneral


class TestConfigParserClients(TestConfigParserBase):

    FILE_PATH = "config.d/clients.yml"
    CONFIG_PARSER_CLASS = ConfigParserClients
    SHORT_DESCR = "Clients config parser"

    def test_valid_cfg(self):
        """{}: valid configuration"""
        self._contains_err()

    def test_unknown_statement(self):
        """{}: unknown statement"""
        self.cfg[0]["test"] = "test"
        self._contains_err("Unknown statement at 'clients' level: 'test'.")

    def test_as(self):
        """{}: AS number"""
        for v in (-1, 0, "test"):
            self.cfg[0]["asn"] = v
            self._contains_err("Error parsing 'asn' at 'clients' level - Invalid ASN: {}.".format(v))

    def test_ip(self):
        """{}: IP address"""
        for v in ("192.0.2.1/24", "2001:db8::1/64"):
            self.cfg[0]["ip"] = v
            self._contains_err("Error parsing 'ip' at 'clients' level - Invalid IP address: {}.".format(v))
        for v in ("192.0.2.1", "10.0.1.1", "1.2.3.4", "fe80::1", "2001:DB8::1"):
            self.cfg[0]["ip"] = v
            self._contains_err()

    def test_duplicate_ip(self):
        """{}: duplicate IP addresses"""
        self.cfg[0]["ip"] = "192.0.2.1"
        self.cfg[1]["ip"] = self.cfg[0]["ip"]
        self._contains_err("Duplicate IP address found: 192.0.2.1.")

    def test_next_hop_policy_no_authorized_addresses_with_list(self):
        """{}: next_hop.policy != authorized_addresses, list given"""
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["policy"] = "strict"
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["authorized_addresses_list"] = ["192.168.0.1"]
        self._contains_err("AS3333 192.0.2.11 is not 'authorized_addresses' but the 'authorized_addresses_list' option is set")

    def test_next_hop_policy_authorized_addresses_empty(self):
        """{}: next_hop.policy authorized_addresses, empty list"""
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["policy"] = "authorized_addresses"
        self._contains_err("set to 'authorized_addresses' but the list of authorized IP addresses ('authorized_addresses_list') is empty")

    def test_next_hop_policy_authorized_addresses_invalid(self):
        """{}: next_hop.policy authorized_addresses, invalid list"""
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["policy"] = "authorized_addresses"
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["authorized_addresses_list"] = "test"
        self._contains_err("Error parsing 'authorized_addresses_list' at 'clients.cfg.filtering.next_hop' level - Invalid format: must be a list.")

    def test_next_hop_policy_authorized_addresses_invalid_address(self):
        """{}: next_hop.policy authorized_addresses, invalid IP addr"""
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["policy"] = "authorized_addresses"
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["authorized_addresses_list"] = ["test"]
        self._contains_err("Error parsing 'authorized_addresses_list' at 'clients.cfg.filtering.next_hop' level - Invalid IP address: test.")

    def test_next_hop_policy_both(self):
        """{}: next_hop.policy and next_hop_policy"""
        self.cfg[0]["cfg"]["filtering"]["next_hop_policy"] = "strict"
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["policy"] = "authorized_addresses"
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["authorized_addresses_list"] = ["test"]
        self._contains_err("Can't use the new 'next_hop' and the old 'next_hop_policy' statements simultaneously")

    def test_next_hop_policy_authorized_addresses_ok(self):
        """{}: next_hop.policy authorized_addresses, ok"""
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["policy"] = "authorized_addresses"
        self.cfg[0]["cfg"]["filtering"]["next_hop"]["authorized_addresses_list"] = ["192.168.0.1"]
        self._contains_err()

    def test_multiple_ip_addresses(self):
        """{}: clients with multiple IP addresses"""
        clients_config = [
            "clients:",
            "  - asn: 111",
            "    ip:",
            "      - '192.0.2.11'",
            "      - '2001:db8:1:1::11'",
            "  - asn: 222",
            "    ip:",
            "      - '192.0.2.21'",
            "      - '2001:db8:1:1::21'",
        ]

        general = ConfigParserGeneral()
        general._load_from_yaml("\n".join([
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
        ]))
        general.parse()

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(clients_config))
        self._contains_err()

        # Duplicate address.
        clients_config = [
            "clients:",
            "  - asn: 111",
            "    ip:",
            "      - '192.0.2.11'",
            "      - '2001:db8:1:1::11'",
            "  - asn: 222",
            "    ip:",
            "      - '192.0.2.11'",
            "      - '2001:db8:1:1::21'",
        ]

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(clients_config))
        self._contains_err("Duplicate IP address found: 192.0.2.11")

    def test_inherit_from_general_cfg(self):
        """{}: inherit from general cfg"""
        general = ConfigParserGeneral()
        general._load_from_yaml("\n".join([
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2"
        ]))
        general.parse()

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join([
            "clients:",
            "  - asn: 111",
            "    ip: 192.0.2.11",
            "  - asn: 222",
            "    ip: 192.0.2.21",
            "    cfg:",
            "      passive: False",
            "      prepend_rs_as: True",
            "      add_path: True",
            "      filtering:",
            "        irrdb:",
            "          enforce_origin_in_as_set: False",
            "          enforce_prefix_in_as_set: False",
            "        rpki:",
            "          enabled: True",
            "          reject_invalid: False",
            "        reject_invalid_as_in_as_path: False",
            "        max_as_path_len: 64",
            "        ipv4_pref_len:",
            "          min: 1",
            "          max: 2",
            "        ipv6_pref_len:",
            "          min: 3",
            "          max: 4",
            "        # test pre v0.6.0 format for next_hop",
            "        next_hop_policy: same-as",
            "        max_prefix:",
            "          peering_db: False",
            "          limit_ipv4: 10",
            "          limit_ipv6: 20",
            "        reject_policy:",
            "          policy: tag",
            "  - asn: 333",
            "    ip: 192.0.2.31",
            "    cfg:",
            "      filtering:",
            "        next_hop:",
            "          policy: 'authorized_addresses'",
            "          authorized_addresses_list:",
            "            - '192.0.2.31'",
            "            - '192.0.2.32'",
            "            - '2001:db8:0:0::31'"
        ]))
        self.cfg.parse()
        self._contains_err()

        client = self.cfg[0]
        self.assertEqual(client["cfg"]["passive"], True)
        self.assertEqual(client["cfg"]["add_path"], False)
        self.assertEqual(client["cfg"]["prepend_rs_as"], False)
        self.assertEqual(client["cfg"]["filtering"]["irrdb"]["enforce_origin_in_as_set"], True)
        self.assertEqual(client["cfg"]["filtering"]["irrdb"]["enforce_prefix_in_as_set"], True)
        self.assertEqual(client["cfg"]["filtering"]["rpki"]["enabled"], False)
        self.assertEqual(client["cfg"]["filtering"]["rpki"]["reject_invalid"], True)
        self.assertEqual(client["cfg"]["filtering"]["reject_invalid_as_in_as_path"], True)
        self.assertEqual(client["cfg"]["filtering"]["max_as_path_len"], 32)
        self.assertEqual(client["cfg"]["filtering"]["ipv4_pref_len"]["min"], 8)
        self.assertEqual(client["cfg"]["filtering"]["ipv4_pref_len"]["max"], 24)
        self.assertEqual(client["cfg"]["filtering"]["ipv6_pref_len"]["min"], 12)
        self.assertEqual(client["cfg"]["filtering"]["ipv6_pref_len"]["max"], 48)
        self.assertEqual(client["cfg"]["filtering"]["next_hop"]["policy"], "strict")
        self.assertEqual(client["cfg"]["filtering"]["max_prefix"]["limit_ipv4"], None)
        self.assertEqual(client["cfg"]["filtering"]["max_prefix"]["limit_ipv6"], None)
        self.assertEqual(client["cfg"]["filtering"]["max_prefix"]["peering_db"], True)
        self.assertEqual(client["cfg"]["filtering"]["reject_policy"]["policy"], "reject")

        client = self.cfg[1]
        self.assertEqual(client["cfg"]["passive"], False)
        self.assertEqual(client["cfg"]["add_path"], True)
        self.assertEqual(client["cfg"]["prepend_rs_as"], True)
        self.assertEqual(client["cfg"]["filtering"]["irrdb"]["enforce_origin_in_as_set"], False)
        self.assertEqual(client["cfg"]["filtering"]["irrdb"]["enforce_prefix_in_as_set"], False)
        self.assertEqual(client["cfg"]["filtering"]["rpki"]["enabled"], True)
        self.assertEqual(client["cfg"]["filtering"]["rpki"]["reject_invalid"], False)
        self.assertEqual(client["cfg"]["filtering"]["reject_invalid_as_in_as_path"], False)
        self.assertEqual(client["cfg"]["filtering"]["max_as_path_len"], 64)
        self.assertEqual(client["cfg"]["filtering"]["ipv4_pref_len"]["min"], 1)
        self.assertEqual(client["cfg"]["filtering"]["ipv4_pref_len"]["max"], 2)
        self.assertEqual(client["cfg"]["filtering"]["ipv6_pref_len"]["min"], 3)
        self.assertEqual(client["cfg"]["filtering"]["ipv6_pref_len"]["max"], 4)
        self.assertEqual(client["cfg"]["filtering"]["next_hop"]["policy"], "same-as")
        self.assertEqual(client["cfg"]["filtering"]["max_prefix"]["limit_ipv4"], 10)
        self.assertEqual(client["cfg"]["filtering"]["max_prefix"]["limit_ipv6"], 20)
        self.assertEqual(client["cfg"]["filtering"]["max_prefix"]["peering_db"], False)
        self.assertEqual(client["cfg"]["filtering"]["reject_policy"]["policy"], "tag")

        client = self.cfg[2]
        self.assertEqual(client["cfg"]["filtering"]["next_hop"]["policy"], "authorized_addresses")
        self.assertEqual(client["cfg"]["filtering"]["next_hop"]["authorized_addresses_list"], [
            "192.0.2.31", "192.0.2.32", "2001:db8:0:0::31"
        ])

    def test_blackhole_filtering_propagation(self):
        """{}: inherit from general cfg - blackhole filtering"""
        clients_config = [
            "clients:",
            "  - asn: 111",
            "    ip: 192.0.2.11",
            "    cfg:",
            "      blackhole_filtering:",             
            "  - asn: 222",
            "    ip: 192.0.2.21",
            "    cfg:",
            "      blackhole_filtering:",
            "        announce_to_client: True",
            "  - asn: 333",
            "    ip: 192.0.2.31",
            "    cfg:",
            "      blackhole_filtering:",
            "        announce_to_client: False",
        ]

        general = ConfigParserGeneral()
        general._load_from_yaml("\n".join([
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
        ]))
        general.parse()

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(clients_config))
        self.cfg.parse()
        self._contains_err()

        self.assertIs(general["blackhole_filtering"]["announce_to_client"], True)

        client = self.cfg[0]
        self.assertIs(client["cfg"]["blackhole_filtering"]["announce_to_client"], True)
        client = self.cfg[1]
        self.assertIs(client["cfg"]["blackhole_filtering"]["announce_to_client"], True)
        client = self.cfg[2]
        self.assertIs(client["cfg"]["blackhole_filtering"]["announce_to_client"], False)

        # ------------------------

        general = ConfigParserGeneral()
        general._load_from_yaml("\n".join([
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  blackhole_filtering:",
            "    announce_to_client: False"  
        ]))
        general.parse()

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(clients_config))
        self.cfg.parse()
        self._contains_err()

        self.assertTrue(general["blackhole_filtering"]["announce_to_client"] is False)

        client = self.cfg[0]
        self.assertIs(client["cfg"]["blackhole_filtering"]["announce_to_client"], False)
        client = self.cfg[1]
        self.assertIs(client["cfg"]["blackhole_filtering"]["announce_to_client"], True)
        client = self.cfg[2]
        self.assertIs(client["cfg"]["blackhole_filtering"]["announce_to_client"], False)

    def test_custom_bgp_communities_ok(self):
        """{}: custom BGP communities"""
        clients_config = [
            "clients:",
            "  - asn: 111",
            "    ip: 192.0.2.11",
            "  - asn: 222",
            "    ip: 192.0.2.21",
            "    cfg:",
            "      attach_custom_communities:"
        ]
        yaml_lines = clients_config + [
            "        - test1",
            "        - test2",
        ]

        general = ConfigParserGeneral()
        general._load_from_yaml("\n".join([
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  custom_communities:",
            "    test1:",
            "      std: '1:1'"
        ]))
        general.parse()

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(yaml_lines))
        self._contains_err("The custom BGP community test2 referenced on client AS222 192.0.2.21 is not declared on the general configuration.")

        yaml_lines = clients_config + [
            "        - test1",
        ]
        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(yaml_lines))
        self._contains_err()

    def test_global_only_option(self):
        """{}: global only option"""
        clients_config = [
            "clients:",
            "  - asn: 222",
            "    ip: 192.0.2.21",
            "    cfg:",
            "      filtering:",
            "        irrdb:",
            "          allow_longer_prefixes: True"
        ]

        general = ConfigParserGeneral()
        general._load_from_yaml("\n".join([
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
        ]))
        general.parse()

        self.cfg = ConfigParserClients(general_cfg=general)
        self.cfg._load_from_yaml("\n".join(clients_config))
        self._contains_err("Unknown statement at 'clients.cfg.filtering.irrdb' level: 'allow_longer_prefixes'")



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys

class InstanceError(Exception):
    pass

class InstanceNotRunning(InstanceError):
    def __init__(self, name, *args, **kwargs):
        InstanceError.__init__(self, *args, **kwargs)
        self.name = name

    def __str__(self):
        return "Instance '{}' is not running.".format(self.name)

class BGPSpeakerInstance(object):
    """This class abstracts a BGP speaker instance.

    Currently, the ``start``, ``stop``, ``is_running`` and
    ``run_cmd`` methods are implemented by the
    :class:`DockerInstance` and :class:`KVMInstance` derived classes,
    while the ``reload_config``, ``get_bgp_session``, ``get_routes``
    and ``log_contains`` methods by the [Docker|KVM]Instance-derived
    :class:`BIRDInstance` class.
    """

    MESSAGE_LOGGING_SUPPORT = True

    DEBUG = False

    @classmethod
    def debug(cls, s):
        if cls.DEBUG or "DEBUG" in os.environ:
            sys.stderr.write("DEBUG: {}\n".format(s))

    def __init__(self, name, ip, mount=[], **kwargs):
        self.name = name
        self.ip = ip
        self.mount = mount

    def set_var_dir(self, var_dir):
        self.var_dir = var_dir

    def get_mounts(self):
        for mount in self.mount:
            res = {}
            res["host"] = mount[0]
            res["container"] = mount[1]
            res["host_filename"] = os.path.split(mount[0])[1]
            res["var_path"] = "{}/{}".format(self.var_dir,
                                             res["host_filename"])
            yield res

    def is_running(self):
        raise NotImplementedError()

    def start(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def reload_config(self):
        raise NotImplementedError()

    def run_cmd(self, args):
        raise NotImplementedError()

    def get_bgp_session(self, other_inst, force_update=False):
        """Get information about the BGP session with ``other_inst``.

        Args:
            other_inst: the
                :class:`BGPSpeakerInstance`
                instance that the current instance is expected to have a
                running BGP session with.

            force_update (bool): if True, the instance must bypass
                any caching mechanism used to keep the BGP sessions status.

        Returns:
            None if the BGP session is not found, otherwise a dictionary
            containing information about the BGP session:

            - "ip": "neighbor IP address",
            - "is_up": [True|False]
        """
        raise NotImplementedError()

    def bgp_session_is_up(self, other_inst, force_update=False):
        """Check if a BGP session with ``other_inst`` is up.

        Args:
            other_inst: the
                :class:`BGPSpeakerInstance`
                instance that the current instance is expected to have a
                running BGP session with.

            force_update (bool): if True, the instance must bypass
                any caching mechanism used to keep the BGP sessions status.

        Returns:
            True if the current instance has a running BGP
            session with ``other_inst``; False otherwise.
        """
        bgp_session_info = self.get_bgp_session(other_inst, force_update)
        if bgp_session_info:
            return bgp_session_info["is_up"]
        raise Exception(
            "Can't get BGP session status for {} on {} "
            "(looking for {})".format(
                other_inst.name, self.name, other_inst.ip
            )
        )

    def get_routes(self, prefix, include_filtered=False, only_best=False):
        """Get a list of all the known routes for ``prefix``.

        Args:
            prefix (str): the IP prefix that returned routes
                must match.

            include_filtered (bool): include filtered routes / rejected
                prefixes in the result.

            only_best (bool): include only the best route toward
                ``prefix``.

        Returns:
            list of :class:`Route` objects.
        """
        raise NotImplementedError()

    def log_contains(self, s):
        """Verifies if the BGP speaker's logs contain the expected message.

        Args:
            s (str): the message that is expected to be found in
                the BGP speaker's logs.

        Returns:
            True or False if the message is found or not.
        """
        raise NotImplementedError()

class Route(object):
    """Details about a route.

    Attributes:

        prefix (str): the IPv4/IPv6 prefix.

        via (str): the IP address of the peer from which the
            route has been received.

        as_path (str): the AS_PATH attribute of the route, in
            the "<asn> <asn> <asn>..." format (example: "1 2 345").

        next_hop (str): the NEXT_HOP attribute of the route.

        filtered (bool): True if the route has been rejected/filtered.

        std_comms (list): list of standard BGP communities (strings in
            the "x:y" format).

        lrg_comms (list): list of large BGP communities (strings in the
            "x:y:z" format).

        ext_comms (list): list of extended BGP communities (strings in
            the "[rt|ro]:x:y" format).

        localpref (int): local-pref.

        reject_reasons (list): list of integers that identify the reasons
            for which the route is considered to be rejected.
    """

    def _parse_bgp_communities(self, communities):
        if not communities:
            return []

        def parse_bgp_community(c):
            v = c
            if v.startswith("("):
                v = v[1:]
            if v.endswith(")"):
                v = v[:-1]
            return ":".join(
                [part.strip() for part in v.split(",")]
            )

        res = []
        for bgp_comm in communities.split(") ("):
            res.append(parse_bgp_community(bgp_comm))
        return res

    def _parse_std_bgp_communities(self, communities):
        return self._parse_bgp_communities(communities)

    def _parse_ext_bgp_communities(self, communities):
        return self._parse_bgp_communities(communities)

    def _parse_lrg_bgp_communities(self, communities):
        return self._parse_bgp_communities(communities)

    def __init__(self, prefix, **kwargs):
        self.prefix = prefix
        self.via = kwargs.get("via", None)
        self.as_path = kwargs.get("as_path", None)
        self.next_hop = kwargs.get("next_hop", None)
        self.localpref = kwargs.get("localpref", None)
        if self.localpref:
            self.localpref = int(self.localpref)
        self.filtered = kwargs.get("filtered", False)
        self.best = kwargs.get("best", None)
        self.std_comms = self._parse_std_bgp_communities(kwargs.get("std_comms", None))
        self.lrg_comms = self._parse_lrg_bgp_communities(kwargs.get("lrg_comms", None))
        self.ext_comms = self._parse_ext_bgp_communities(kwargs.get("ext_comms", None))
        self.reject_reasons = []

    def process_reject_cause(self, re_pattern):
        # If a route is marked to be rejected using the 'reject_cause'
        # community it must be also set with LOCAL_PREF == 0.
        if self.localpref != 1:
            return

        reject_cause_zero_found = False
        reasons = []

        # Iterating over the original list (from which the 'reject_cause'
        # community will be removed) and its copy (to keep consistent the
        # set of values over which iterate).
        for orig_list, dup_list in [(self.std_comms, list(self.std_comms)),
                                    (self.lrg_comms, list(self.lrg_comms)),
                                    (self.ext_comms, list(self.ext_comms))]:
            for comm in dup_list:
                match = re_pattern.match(comm)

                if not match:
                    continue

                reason = int(match.group(1))

                if reason == 0:
                    reject_cause_zero_found = True
                else:
                    reasons.append(reason)

                orig_list.remove(comm)

        if reject_cause_zero_found:
            self.reject_reasons = reasons
            self.filtered = True

    def __str__(self):
        return str({
            "prefix": self.prefix,
            "via": self.via,
            "as_path": self.as_path,
            "next_hop": self.next_hop,
            "localpref": self.localpref,
            "filtered": self.filtered,
            "reject_reasons": ", ".join(map(str, self.reject_reasons)),
            "best": self.best,
            "std_comms": self.std_comms,
            "lrg_comms": self.lrg_comms,
            "ext_comms": self.ext_comms,
        })

    def __repr__(self):
        return self.__str__()



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
try:
    # For Python 3.0 and later
    from urllib.request import urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen

from .errors import IXFDBError, IXFDBSchemaError

class IXFDB(object):

    def __init__(self, input_object="https://db.ix-f.net/api/ixp"):
        self.raw_data = None

        if isinstance(input_object, dict):
            self.raw_data = input_object
        elif isinstance(input_object, file):
            raw = input_object.read()
        else:
            try:
                response = urlopen(input_object)
                raw = response.read().decode("utf-8")
            except Exception as e:
                raise IXFDBError(
                    "Error while retrieving IX-F DB "
                    "JSON file from {}: {}".format(
                        input_object, str(e)
                    )
                )

        if not self.raw_data:
            try:
                self.raw_data = json.loads(raw)
            except Exception as e:
                raise EuroIXSchemaError(
                    "Error while processing JSON data: {}".format(str(e))
                )

        if "data" not in self.raw_data:
            raise IXFDBSchemaError(
                "Missing 'data' element"
            )

        if not isinstance(self.raw_data["data"], list):
            raise IXFDBSchemaError(
                "The 'data' element is not a list."
            )

        self.ixp_list = self.raw_data["data"]



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ipaddr
import logging
import re

from .base import BaseConfigEnricher, BaseConfigEnricherThread
from ..errors import BuilderError, ARouteServerError
from ..irrdb import ASSet, RSet, IRRDBTools


class IRRDBConfigEnricher_WorkerThread(BaseConfigEnricherThread):

    TARGET_FIELD = None

    def __init__(self, *args, **kwargs):
        BaseConfigEnricherThread.__init__(self, *args, **kwargs)

        self.ip_ver = None
        self.irrdbtools_cfg = None

    def do_task(self, task):
        _, dest_descr, as_set_name = task
        data = self._get_external_data(dest_descr, as_set_name)
        return data

    def save_data(self, task, data):
        as_set, _, _ = task
        as_set[self.TARGET_FIELD].extend(
            [_ for _ in data if _ not in as_set[self.TARGET_FIELD]]
        )

class IRRDBConfigEnricher_WorkerThread_Prefixes(IRRDBConfigEnricher_WorkerThread):

    DESCR = "IRRdb prefixes"
    TARGET_FIELD = "prefixes"

    def _get_external_data(self, dest_descr, as_set_name):
        errors = False
        ip_versions = [self.ip_ver] if self.ip_ver else [4, 6]
        res = []
        for ip_ver in ip_versions:
            try:
                prefixes = RSet(as_set_name, ip_ver,
                                **self.irrdbtools_cfg).prefixes
                if not prefixes:
                    logging.warning("No IPv{} prefixes found in "
                                    "{} for {}".format(
                                        ip_ver, as_set_name, dest_descr))
                res.extend(prefixes)
            except ARouteServerError as e:
                errors = True
                logging.error(
                    "Error while retrieving r_set "
                    "{} for {} IPv{}: {}".format(
                        as_set_name, dest_descr, ip_ver, str(e)
                    )
                )

        if errors:
            raise BuilderError()

        return res

class IRRDBConfigEnricher_WorkerThread_OriginASNs(IRRDBConfigEnricher_WorkerThread):

    DESCR = "IRRdb origin ASNs"
    TARGET_FIELD = "asns"

    def _get_external_data(self, dest_descr, as_set_name):
        errors = False
        try:
            asns = ASSet(as_set_name, **self.irrdbtools_cfg).asns
            if not asns:
                logging.warning("No origin ASNs found in "
                                "{} for {}".format(
                                    as_set_name, dest_descr))
            return asns
        except ARouteServerError as e:
            errors = True
            logging.error(
                "Error while retrieving as_set {} for {}: {}".format(
                    as_set_name, dest_descr, str(e)
                )
            )

        if errors:
            raise BuilderError()

class IRRDBConfigEnricher(BaseConfigEnricher):

    WORKER_THREAD_CLASS = None

    def prepare(self):
        if self.builder.as_sets is not None:
            return

        as_sets = []
        errors = False

        def normalize_as_set_id(s):
            return re.sub("[^a-zA-Z0-9_]", "_", s)

        def get_as_set_by_name(name, used_by_client=None):
            for as_set in as_sets:
                if as_set["name"] == name:
                    if used_by_client:
                        as_set["used_by"].append(used_by_client)
                    return as_set
            return None

        def use_as_set(as_set_name, used_by_client=None):
            # Returns: id of the added/existing as_set
            existing = get_as_set_by_name(as_set_name)
            if existing:
                if used_by_client:
                    existing["used_by"].append(used_by_client)
                return existing["id"]

            new_as_set_id = normalize_as_set_id(as_set_name)
            as_sets.append({
                "id": new_as_set_id,
                "name": as_set_name,
                "asns": [],
                "prefixes": [],
                "used_by": [used_by_client] if used_by_client else []
            })
            return new_as_set_id

        # Add to as_sets all the AS-SETs reported in the 'asns' section.
        for asn in self.builder.cfg_asns.cfg["asns"]:
            self.builder.cfg_asns[asn]["as_set_ids"] = []

            if not self.builder.cfg_asns[asn]["as_sets"]:
                continue

            for as_set in self.builder.cfg_asns[asn]["as_sets"]:
                self.builder.cfg_asns[asn]["as_set_ids"].append(
                    use_as_set(as_set)
                )

        # Add to as_sets all the AS-SETs reported in the 'clients' section.
        for client in self.builder.cfg_clients.cfg["clients"]:
            client_irrdb = client["cfg"]["filtering"]["irrdb"]
            client_irrdb["as_set_ids"] = []

            if not client_irrdb["enforce_origin_in_as_set"] and \
                not client_irrdb["enforce_prefix_in_as_set"] and \
                not self.builder.cfg_general["filtering"]["irrdb"]["tag_as_set"]:
                    # Client does not require AS-SETs info to be gathered.
                    continue

            if self.builder.ip_ver is not None:
                ip = client["ip"]
                if ipaddr.IPAddress(ip).version != self.builder.ip_ver:
                    # The address family of this client is not the
                    # current one used to build the configuration.
                    continue

            if client_irrdb["as_sets"]:
                # Client has its own specific set of AS-SETs.
                for as_set in client_irrdb["as_sets"]:
                    client_irrdb["as_set_ids"].append(
                        use_as_set(as_set, "client {}".format(client["id"]))
                    )
                continue

            # Client needs AS-SETs info but has not its own set of AS-SETs.

            # If client's ASN is configured in the 'asns' section and has
            # one or more AS-SETs, the client configuration will be based
            # on those.
            asn = "AS{}".format(client["asn"])
            if asn in self.builder.cfg_asns.cfg["asns"] and \
                self.builder.cfg_asns.cfg["asns"][asn]["as_sets"]:
                for as_set in self.builder.cfg_asns.cfg["asns"][asn]["as_sets"]:
                    client_irrdb["as_set_ids"].append(
                        use_as_set(as_set, "client {}".format(client["id"]))
                    )
                continue

            # No AS-SETs found for the client's ASN in the 'asns' section.
            logging.warning("No AS-SET provided for the '{}' client. "
                            "Only AS{} will be expanded.".format(
                                client["id"], client["asn"]
                            ))

            client_irrdb["as_set_ids"].append(
                use_as_set("AS{}".format(client["asn"]),
                           "client {}".format(client["id"]))
            )

        # Removing unreferenced AS-SETs.
        for as_set in as_sets:
            if not as_set["used_by"]:
                logging.debug("Removing unreferenced AS-SET: "
                              "{}".format(as_set["name"]))
        as_sets = [as_set for as_set in as_sets if as_set["used_by"]]

        self.builder.as_sets = {}
        for as_set in as_sets:
            self.builder.as_sets[as_set["id"]] = as_set

    def _config_thread(self, thread):
        thread.ip_ver = self.builder.ip_ver
        thread.irrdbtools_cfg = {
            "bgpq3_path": self.builder.bgpq3_path,
            "bgpq3_host": self.builder.bgpq3_host,
            "bgpq3_sources": self.builder.bgpq3_sources,
            "cache_dir": self.builder.cache_dir,
            "cache_expiry": self.builder.cache_expiry,
        }

    def add_tasks(self):
        # Enqueuing tasks.
        for as_set_id, as_set in self.builder.as_sets.items():
            used_by = ", ".join(as_set["used_by"])
            self.tasks_q.put((as_set, used_by, as_set["name"]))

class IRRDBConfigEnricher_OriginASNs(IRRDBConfigEnricher):

    WORKER_THREAD_CLASS = IRRDBConfigEnricher_WorkerThread_OriginASNs

class IRRDBConfigEnricher_Prefixes(IRRDBConfigEnricher):

    WORKER_THREAD_CLASS = IRRDBConfigEnricher_WorkerThread_Prefixes

    def _config_thread(self, thread):
        IRRDBConfigEnricher._config_thread(self, thread)
        thread.irrdbtools_cfg["allow_longer_prefixes"] = \
            self.builder.cfg_general["filtering"]["irrdb"]["allow_longer_prefixes"]



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ipaddr
import unittest

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDInstance, \
                                                          OpenBGPD60Instance
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance

class BasicScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    AS_SET = {
        "AS-AS1": [1],
        "AS-AS1_CUSTOMERS": [101],
        "AS-AS2": [2],
        "AS-AS2_CUSTOMERS": [101],
    }
    R_SET = {
        "AS-AS1": [
            "AS1_allowed_prefixes",
            "pref_len1"
        ],
        "AS-AS1_CUSTOMERS": [
            "AS101_allowed_prefixes"
        ],
        "AS-AS2": [
            "AS2_allowed_prefixes"
        ],
        "AS-AS2_CUSTOMERS": [
            "AS101_allowed_prefixes"
        ],
    }

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            # Run AS3 as soon as possible because it's configured with passive
            # session, so while other instance come up rs has more time to
            # setup the session with it.
            cls.CLIENT_INSTANCE_CLASS(
                "AS3",
                cls.DATA["AS3_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS3.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS1_1",
                cls.DATA["AS1_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1_1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS1_2",
                cls.DATA["AS1_2_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1_2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS101",
                cls.DATA["AS101_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS101.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            )
        ]

    def set_instance_variables(self):
        self.AS1_1 = self._get_instance_by_name("AS1_1")
        self.AS1_2 = self._get_instance_by_name("AS1_2")
        self.AS2 = self._get_instance_by_name("AS2")
        # AS3 is passive, rs client configured with passive: False
        self.AS3 = self._get_instance_by_name("AS3")
        self.AS101 = self._get_instance_by_name("AS101")
        self.rs = self._get_instance_by_name("rs")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1_1)
        self.session_is_up(self.rs, self.AS1_2)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS3)
        self.session_is_up(self.AS101, self.AS1_1)
        self.session_is_up(self.AS101, self.AS1_2)
        self.session_is_up(self.AS101, self.AS2)

    def test_021_session_configured_via_local_files(self):
        """{}: session configured via local include files"""

        # A dummy session is configured using local include files.
        # The following tests if those files are really included.
        self.session_exists(self.rs, self.DATA["RoutesCollector_IPAddress"])

    def test_030_good_prefixes_received_by_rs(self):
        """{}: good prefixes received by rs"""
        if isinstance(self.rs, BIRDInstance):
            ext_comm_rpki_unknown = ["generic:0x43000000:0x1"]
        else:
            ext_comm_rpki_unknown = []
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_1,
                           next_hop=self.AS1_1, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS1_good2"], self.AS1_1,
                           next_hop=self.AS1_1, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_2,
                           next_hop=self.AS1_2, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS1_good2"], self.AS1_2,
                           next_hop=self.AS1_2, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        # AS1_good3 is announced by AS1_2 with NEXT_HOP = AS1_1
        self.receive_route(self.rs, self.DATA["AS1_good3"], self.AS1_2,
                           next_hop=self.AS1_1, as_path="1",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS2,
                           next_hop=self.AS2, as_path="2",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS2_good2"], self.AS2,
                           next_hop=self.AS2, as_path="2",
                           std_comms=[], ext_comms=ext_comm_rpki_unknown, lrg_comms=[])

        # rs should not receive prefixes with the following criteria
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS2)
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS1_good2"], self.AS2)
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS1_1)
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.rs, self.DATA["AS2_good2"], self.AS1_1)

        # AS_PATH should match the expectations
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_1, as_path="1")
        self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS2, as_path="2")

        # NEXT_HOP should match the expectations
        self.receive_route(self.rs, self.DATA["AS1_good1"], self.AS1_1, as_path="1", next_hop=self.AS1_1)
        self.receive_route(self.rs, self.DATA["AS2_good1"], self.AS2, as_path="2", next_hop=self.AS2)

    def test_030_good_prefixes_received_by_rs_nonclient_nexthop(self):
        """{}: good prefixes received by rs: non-client NEXT_HOP"""

        self.receive_route(self.rs, self.DATA["AS2_nonclient_nexthop1"],
                           as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop1_nh"])

    def test_040_bad_prefixes_received_by_rs_bogon(self):
        """{}: bad prefixes received by rs: bogon"""
        self.receive_route(self.rs, self.DATA["bogon1"], self.AS1_1,
                           as_path="1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=2)
        self.log_contains(self.rs, "prefix is bogon - REJECTING " + self.DATA["bogon1"])

    def test_040_bad_prefixes_received_by_rs_global_blacklist(self):
        """{}: bad prefixes received by rs: global blacklist"""

        self.receive_route(self.rs, self.DATA["local1"], self.AS1_1,
                           as_path="1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=3)
        self.log_contains(self.rs, "prefix is in global blacklist - REJECTING " + self.DATA["local1"])

    def test_040_bad_prefixes_received_by_rs_prefix_len(self):
        """{}: bad prefixes received by rs: invalid prefix-len"""

        ip_ver = ipaddr.IPNetwork(self.DATA["pref_len1"]).version

        self.receive_route(self.rs, self.DATA["pref_len1"], self.AS1_1,
                           as_path="1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=13)
        pref_len = 7 if ip_ver == 4 else 16
        pref_len_range = "8-24" if ip_ver == 4 else "17-48"
        self.log_contains(self.rs, "prefix len [{}] not in {} - REJECTING {}".format(
            pref_len, pref_len_range, self.DATA["pref_len1"]))

    def test_040_bad_prefixes_received_by_rs_leftmost_asn(self):
        """{}: bad prefixes received by rs: left-most ASN"""

        self.receive_route(self.rs, self.DATA["peer_as1"], self.AS1_1,
                           as_path="2 1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=6)
        self.log_contains(self.rs, "invalid left-most ASN [2] - REJECTING " + self.DATA["peer_as1"])

    def test_040_bad_prefixes_received_by_rs_invalid_aspath(self):
        """{}: bad prefixes received by rs: invalid ASN in AS-PATH"""

        self.receive_route(self.rs, self.DATA["invalid_asn1"], self.AS1_1,
                           as_path="1 65536 1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=7)
        self.log_contains(self.rs, "AS_PATH [(path 1 65536 1)] contains invalid ASN - REJECTING " + self.DATA["invalid_asn1"])

    def test_040_bad_prefixes_received_by_rs_transitfree_as_path(self):
        """{}: bad prefixes received by rs: transit-free ASN in AS-PATH"""

        self.receive_route(self.rs, self.DATA["AS101_transitfree_1"],
                           self.AS1_1, as_path="1 101 174",
                           next_hop=self.AS1_1, filtered=True, reject_reason=8)
        self.log_contains(self.rs, "AS_PATH [(path 1 101 174)] contains transit-free ASN - REJECTING " + self.DATA["AS101_transitfree_1"])

    def test_040_bad_prefixes_received_by_rs_aspath_len(self):
        """{}: bad prefixes received by rs: AS_PATH len"""

        self.receive_route(self.rs, self.DATA["aspath_len1"], self.AS1_1,
                           as_path="1 2 2 2 2 2 2 1", next_hop=self.AS1_1,
                           filtered=True, reject_reason=1)
        self.log_contains(self.rs, "AS_PATH len [8] longer than 6 - REJECTING " + self.DATA["aspath_len1"])

    def test_040_bad_prefixes_received_by_rs_client_blacklist(self):
        """{}: bad prefixes received by rs: client blacklist"""

        self.receive_route(self.rs, self.DATA["AS3_blacklist1"], self.AS3,
                           as_path="3", next_hop=self.AS3,
                           filtered=True, reject_reason=11)
        self.log_contains(self.rs, "prefix is in client's blacklist - REJECTING " + self.DATA["AS3_blacklist1"])

    def test_040_bad_prefixes_received_by_rs_invalid_nexthop(self):
        """{}: bad prefixes received by rs: invalid NEXT_HOP"""

        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS1_2,
                           as_path="1 101", next_hop=self.AS101,
                           filtered=True, reject_reason=5)
        self.log_contains(self.rs, "NEXT_HOP [" + self.AS101.ip + "] not allowed - REJECTING " + self.DATA["AS101_good1"])

    def test_040_bad_prefixes_received_by_rs_unknown_nonclient_nexthop(self):
        """{}: bad prefixes received by rs: unknown NEXT_HOP"""

        self.receive_route(self.rs, self.DATA["AS2_nonclient_nexthop2"],
                           self.AS2, as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop2_nh"],
                           filtered=True, reject_reason=5)
        self.log_contains(self.rs, "NEXT_HOP [" + self.DATA["AS2_nonclient_nexthop2_nh"] + "] not allowed - REJECTING " + self.DATA["AS2_nonclient_nexthop2"])

    def test_040_bad_prefixes_received_by_rs_no_rset(self):
        """{}: bad prefixes received by rs: prefix not in AS-SET"""

        # AS101_no_rset is not included in AS-AS1_CUSTOMERS nor in AS-AS2_CUSTOMERS, so it's
        # rejected by the rs.
        self.receive_route(self.rs, self.DATA["AS101_no_rset"], self.AS1_1,
                           as_path="1 101", next_hop=self.AS1_1,
                           filtered=True, reject_reason=12)
        self.receive_route(self.rs, self.DATA["AS101_no_rset"], self.AS2,
                           as_path="2 101", next_hop=self.AS2,
                           filtered=True, reject_reason=12)
        self.log_contains(self.rs, "prefix not in client's r_set - REJECTING " + self.DATA["AS101_no_rset"])

    def test_040_bad_prefixes_received_by_rs_no_asset(self):
        """{}: bad prefixes received by rs: origin not in AS-SET"""

        # AS102_no_asset is announced by an (hypothetical) AS102 to AS101,
        # and AS102 is not included in AS-AS1_CUSTOMERS nor in AS-AS2_CUSTOMERS, so it's
        # rejected by the rs.
        self.receive_route(self.rs, self.DATA["AS102_no_asset"], self.AS1_1,
                           as_path="1 101 102", next_hop=self.AS1_1,
                           filtered=True, reject_reason=9)
        self.receive_route(self.rs, self.DATA["AS102_no_asset"], self.AS2,
                           as_path="2 101 102", next_hop=self.AS2,
                           filtered=True, reject_reason=9)
        self.log_contains(self.rs, "origin ASN [102] not in allowed as-sets - REJECTING " + self.DATA["AS102_no_asset"])

    def test_040_bad_prefix_not_ipv6_global_unicat(self):
        """{}: bad prefixes received by rs: not IPv6 global unicast space"""
        if "AS101_no_ipv6_gl_uni" not in self.DATA:
            # it's an IPv4-only scenario
            return
        self.receive_route(self.rs, self.DATA["AS101_no_ipv6_gl_uni"],
                           filtered=True, reject_reason=10)
        self.log_contains(self.rs, "prefix is not in IPv6 Global Unicast space - REJECTING " + self.DATA["AS101_no_ipv6_gl_uni"])

    def test_040_default_rejected_by_rs(self):
        """{}: bad prefixes received by rs: default route"""
        self.receive_route(self.rs, self.DATA["Default_route"],
                           other_inst=self.AS3,
                           filtered=True, reject_reason=(2, 10))
        if ipaddr.IPNetwork(self.DATA["Default_route"]).version == 4:
            msg = "prefix is bogon - REJECTING " + self.DATA["Default_route"]
        else:
            msg = "prefix is not in IPv6 Global Unicast space - REJECTING " + self.DATA["Default_route"]
        self.log_contains(self.rs, msg)

    def test_041_bad_prefixes_not_received_by_clients(self):
        """{}: bad prefixes not received by clients"""
        for prefix in (self.DATA["bogon1"],
                       self.DATA["local1"],
                       self.DATA["pref_len1"],
                       self.DATA["peer_as1"],
                       self.DATA["invalid_asn1"],
                       self.DATA["aspath_len1"],
                       self.DATA["AS2_nonclient_nexthop2"]):
            for inst in (self.AS2, self.AS3):
                with self.assertRaisesRegexp(AssertionError, "Routes not found."):
                    self.receive_route(inst, prefix)

        # Among the clients, only AS3 is expected to not see the 
        # following prefixes because AS1 and AS2
        # receive them on their session with AS101
        for prefix in (self.DATA["AS101_no_rset"],
                       self.DATA["AS102_no_asset"]):
            with self.assertRaisesRegexp(AssertionError, "Routes not found."):
                self.receive_route(self.AS3, prefix)

    def test_045_rpki_valid_prefix(self):
        """{}: RPKI, valid prefix received by rs"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("RPKI not supported by OpenBGPD")

        self.receive_route(self.rs, self.DATA["AS101_roa_valid1"], self.AS1_1,
                           as_path="1 101")
        self.receive_route(self.rs, self.DATA["AS101_roa_valid1"], self.AS2,
                           as_path="2 101")

    def test_045_rpki_invalid_prefix_asn(self):
        """{}: RPKI, invalid prefix (bad ASN) received by rs"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("RPKI not supported by OpenBGPD")

        self.receive_route(self.rs, self.DATA["AS101_roa_invalid1"], self.AS1_1,
                           as_path="1 101", filtered=True,
                           ext_comms=["generic:0x43000000:0x2"])
        self.receive_route(self.rs, self.DATA["AS101_roa_invalid1"], self.AS2,
                           as_path="2 101", filtered=True,
                           ext_comms=["generic:0x43000000:0x2"])

    def test_045_rpki_invalid_prefix_length(self):
        """{}: RPKI, invalid prefix (bad length) received by rs"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("RPKI not supported by OpenBGPD")

        self.receive_route(self.rs, self.DATA["AS101_roa_badlen"], self.AS1_1,
                           as_path="1 101", filtered=True,
                           ext_comms=["generic:0x43000000:0x2"])
        self.receive_route(self.rs, self.DATA["AS101_roa_badlen"], self.AS2,
                           as_path="2 101", filtered=True,
                           ext_comms=["generic:0x43000000:0x2"])

    def test_045_rpki_valid_prefix_propagated_to_clients(self):
        """{}: RPKI, valid prefix propagated to clients"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("RPKI not supported by OpenBGPD")

        self.receive_route(self.AS3, self.DATA["AS101_roa_valid1"], self.rs)

    def test_045_rpki_invalid_prefixes_not_propagated_to_clients(self):
        """{}: RPKI, invalid prefix (bad ASN) not propagated to clients"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("RPKI not supported by OpenBGPD")

        for pref_id in ("AS101_roa_invalid1", "AS101_roa_badlen"):
            with self.assertRaisesRegexp(AssertionError, "Routes not found."):
                self.receive_route(self.AS3, self.DATA[pref_id])

    def test_045_blackhole_with_roa(self):
        """{}: RPKI, blackhole request for a covered prefix"""
        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("RPKI not supported by OpenBGPD")

        self.receive_route(self.rs, self.DATA["AS101_roa_blackhole"], self.AS1_1, as_path="1 101",
                           std_comms=["65535:666"], lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS101_roa_blackhole"], self.AS2, as_path="2 101",
                           std_comms=["65535:666"], lrg_comms=[])
        for inst in (self.AS1_1, self.AS2):
            self.log_contains(self.rs, "blackhole filtering request from {{inst}} - ACCEPTING {}".format(
                self.DATA["AS101_roa_blackhole"]), {"inst": inst})
        self.receive_route(self.AS3, self.DATA["AS101_roa_blackhole"], self.rs,
                           next_hop=self.DATA["blackhole_IP"],
                           std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_050_prefixes_from_AS101_received_by_its_upstreams(self):
        """{}: prefixes from AS101 received by its upstreams"""
        self.receive_route(self.AS1_1, self.DATA["AS101_good1"], self.AS101)
        self.receive_route(self.AS1_2, self.DATA["AS101_good1"], self.AS101)
        self.receive_route(self.AS2, self.DATA["AS101_good1"], self.AS101)

    def test_051_prefixes_from_AS101_received_by_rs(self):
        """{}: prefixes from AS101 received by rs"""

        # rs should receive these prefixes that AS101 announces to AS1 and AS2.
        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS1_1, as_path="1 101", next_hop=self.AS1_1)
        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS2, as_path="2 101", next_hop=self.AS2)

        # AS101 peers with AS1_2 on the same network of rs
        # and AS1_2 has not 'next-hop-self' on the session with rs:
        # next-hop received by rs is the one of AS101, so the prefix
        # should be filtered by the rs.
        self.receive_route(self.rs, self.DATA["AS101_good1"], self.AS1_2,
                           as_path="1 101", next_hop=self.AS101,
                           filtered=True, reject_reason=5)

    def test_060_communities_as_seen_by_AS101_upstreams(self):
        """{}: bad communities as seen by AS101 upstreams"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route(inst, self.DATA["AS101_bad_std_comm"], self.AS101, std_comms=["65530:0"])
            self.receive_route(inst, self.DATA["AS101_bad_lrg_comm"], self.AS101, lrg_comms=["999:65530:0"])
            self.receive_route(inst, self.DATA["AS101_other_s_comm"], self.AS101, std_comms=["888:0"])
            self.receive_route(inst, self.DATA["AS101_other_l_comm"], self.AS101, lrg_comms=["888:0:0"])
            self.receive_route(inst, self.DATA["AS101_bad_good_comms"], self.AS101, std_comms=["777:0", "65530:1"], lrg_comms=["777:0:0", "999:65530:1"])

    def test_061_bad_communities_scrubbed_by_rs_std(self):
        """{}: bad communities scrubbed by rs (std)"""
        
        self.receive_route(self.rs, self.DATA["AS101_bad_std_comm"], self.AS1_1, std_comms=[])
        self.receive_route(self.rs, self.DATA["AS101_bad_std_comm"], self.AS2, std_comms=[])

    def test_061_bad_communities_scrubbed_by_rs_lrg(self):
        """{}: bad communities scrubbed by rs (lrg)"""
        if isinstance(self.rs, OpenBGPD60Instance):
            raise unittest.SkipTest("Large comms not supported by OpenBGPD 6.0")

        self.receive_route(self.rs, self.DATA["AS101_bad_lrg_comm"], self.AS1_1, lrg_comms=[])
        self.receive_route(self.rs, self.DATA["AS101_bad_lrg_comm"], self.AS2, lrg_comms=[])

    def test_062_other_communities_not_scrubbed_by_rs_std(self):
        """{}: other communities not scrubbed by rs (std)"""

        for inst in (self.AS1_1, self.AS2):
            self.receive_route(self.rs, self.DATA["AS101_other_s_comm"], inst, std_comms=["888:0"])
            self.receive_route(self.rs, self.DATA["AS101_bad_good_comms"], inst, std_comms=["777:0"])

    def test_062_other_communities_not_scrubbed_by_rs_lrg(self):
        """{}: other communities not scrubbed by rs (lrg)"""
        if isinstance(self.rs, OpenBGPD60Instance):
            raise unittest.SkipTest("Large comms not supported by OpenBGPD 6.0")

        for inst in (self.AS1_1, self.AS2):
            self.receive_route(self.rs, self.DATA["AS101_other_l_comm"], inst, lrg_comms=["888:0:0"])
            self.receive_route(self.rs, self.DATA["AS101_bad_good_comms"], inst, lrg_comms=["777:0:0"])

    def test_070_blackhole_filtering_as_seen_by_rs_BLACKHOLE(self):
        """{}: blackhole filtering requests as seen by rs (BLACKHOLE)"""

        self.receive_route(self.rs, self.DATA["AS2_blackhole1"], self.AS2, next_hop=self.AS2, as_path="2",
                           std_comms=["65535:666"], lrg_comms=[])
        self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole1"], {"AS2_1": self.AS2})

    def test_070_blackhole_filtering_as_seen_by_rs_std_cust(self):
        """{}: blackhole filtering requests as seen by rs (std cust)"""

        self.receive_route(self.rs, self.DATA["AS2_blackhole2"], self.AS2, next_hop=self.AS2, as_path="2",
                           std_comms=["65534:0"], lrg_comms=[])
        self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole2"], {"AS2_1": self.AS2})

    def test_070_blackhole_filtering_as_seen_by_rs_lrg_cust(self):
        """{}: blackhole filtering requests as seen by rs (lrg cust)"""
        if isinstance(self.rs, OpenBGPD60Instance):
            raise unittest.SkipTest("Large comms not supported by OpenBGPD 6.0")

        self.receive_route(self.rs, self.DATA["AS2_blackhole3"], self.AS2, next_hop=self.AS2, as_path="2",
                           std_comms=[], lrg_comms=["65534:0:0"])
        self.log_contains(self.rs, "blackhole filtering request from {AS2_1} - ACCEPTING " + self.DATA["AS2_blackhole3"], {"AS2_1": self.AS2})

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients_BLACKHOLE(self):
        """{}: blackholed prefixes as seen by enabled clients (BLACKHOLE)"""
        if isinstance(self.rs, OpenBGPD60Instance) and ":" in self.DATA["AS2_blackhole1"]:
            # OpenBGPD < 6.1 bug: https://github.com/pierky/arouteserver/issues/3
            # fixed by https://github.com/openbsd/src/commit/f1385c8f4f9b9e193ff65d9f2039862d3e230a45
            raise unittest.SkipTest("Not working on OpenBGPD 6.0")

        for inst in (self.AS1_1, self.AS3):
            self.receive_route(inst, self.DATA["AS2_blackhole1"], self.rs, next_hop=self.DATA["blackhole_IP"],
                               std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients_std_cust(self):
        """{}: blackholed prefixes as seen by enabled clients (std_cust)"""
        if isinstance(self.rs, OpenBGPD60Instance) and ":" in self.DATA["AS2_blackhole1"]:
            # OpenBGPD < 6.1 bug: https://github.com/pierky/arouteserver/issues/3
            # fixed by https://github.com/openbsd/src/commit/f1385c8f4f9b9e193ff65d9f2039862d3e230a45
            raise unittest.SkipTest("Not working on OpenBGPD 6.0")

        for inst in (self.AS1_1, self.AS3):
            self.receive_route(inst, self.DATA["AS2_blackhole2"], self.rs, next_hop=self.DATA["blackhole_IP"],
                               std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_071_blackholed_prefixes_as_seen_by_enabled_clients_lrg_cust(self):
        """{}: blackholed prefixes as seen by enabled clients (lrg_cust)"""
        if isinstance(self.rs, OpenBGPD60Instance):
            raise unittest.SkipTest("Large comms not supported by OpenBGPD 6.0")

        for inst in (self.AS1_1, self.AS3):
            self.receive_route(inst, self.DATA["AS2_blackhole3"], self.rs, next_hop=self.DATA["blackhole_IP"],
                               std_comms=["65535:666", "65535:65281"], lrg_comms=[])

    def test_071_blackholed_prefixes_triggered_via_unsupported_lrg_comms(self):
        """{}: blackholed prefixes triggered via unsupported lrg comms"""
        if not isinstance(self.rs, OpenBGPD60Instance):
            # Test valid only for BGP daemons that don't support large comms
            raise unittest.SkipTest("Large comms are supported here")

        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS1_1, self.DATA["AS2_blackhole3"])
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS3, self.DATA["AS2_blackhole3"])

    def test_071_blackholed_prefixes_not_seen_by_not_enabled_clients(self):
        """{}: blackholed prefixes not seen by not enabled clients"""

        # AS1_2 not enabled to receive blackholed prefixes
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS1_2, self.DATA["AS2_blackhole1"])
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS1_2, self.DATA["AS2_blackhole2"])
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS1_2, self.DATA["AS2_blackhole3"])
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole1"]), {"AS1_2": self.AS1_2})
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole2"]), {"AS1_2": self.AS1_2})
        self.log_contains(self.rs, "client {{AS1_2}} not enabled to receive blackhole prefixes - NOT ANNOUNCING {pref} TO {{AS1_2}}".format(pref=self.DATA["AS2_blackhole3"]), {"AS1_2": self.AS1_2})

    def test_080_control_communities_AS1_only(self):
        """{}: control communities, announce to AS1 only"""

        self.receive_route(self.AS1_1, self.DATA["AS3_cc_AS1only"], self.rs,
                           as_path="3", next_hop=self.AS3,
                           std_comms=[], lrg_comms=[])
        self.receive_route(self.AS1_2, self.DATA["AS3_cc_AS1only"], self.rs,
                           as_path="3", next_hop=self.AS3,
                           std_comms=[], lrg_comms=[])

        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS2, self.DATA["AS3_cc_AS1only"])
        self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{AS2}}".format(self.DATA["AS3_cc_AS1only"]), {"AS2": self.AS2})

    def test_080_control_communities_not_AS1(self):
        """{}: control communities, announce to all except AS1"""

        self.receive_route(self.AS2, self.DATA["AS3_cc_not_AS1"], self.rs,
                           as_path="3", next_hop=self.AS3,
                           std_comms=[], lrg_comms=[])

        for inst in (self.AS1_1, self.AS1_2):
            with self.assertRaisesRegexp(AssertionError, "Routes not found."):
                self.receive_route(inst, self.DATA["AS3_cc_not_AS1"])
            self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{other_inst}}".format(self.DATA["AS3_cc_not_AS1"]), {"other_inst": inst})

    def test_080_control_communities_none(self):
        """{}: control communities, don't announce to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            with self.assertRaisesRegexp(AssertionError, "Routes not found."):
                self.receive_route(inst, self.DATA["AS3_cc_none"])
            self.log_contains(self.rs, "route didn't pass control communities checks - NOT ANNOUNCING {} TO {{other_inst}}".format(self.DATA["AS3_cc_none"]), {"other_inst": inst})

    def test_081_control_communities_prepend1any(self):
        """{}: control communities, prepend once to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route(inst, self.DATA["AS3_prepend1any"], self.rs, as_path="3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_081_control_communities_prepend2any(self):
        """{}: control communities, prepend twice to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route(inst, self.DATA["AS3_prepend2any"], self.rs, as_path="3 3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_081_control_communities_prepend3any(self):
        """{}: control communities, prepend thrice to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route(inst, self.DATA["AS3_prepend3any"], self.rs, as_path="3 3 3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_082_control_communities_prepend1_AS1(self):
        """{}: control communities, prepend once to AS1"""

        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, self.DATA["AS3_prepend1_AS1"], self.rs, as_path="3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])
        self.receive_route(self.AS2, self.DATA["AS3_prepend1_AS1"], self.rs, as_path="3",
                           next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_082_control_communities_prepend2_AS2(self):
        """{}: control communities, prepend twice to AS2"""

        self.receive_route(self.AS2, self.DATA["AS3_prepend2_AS2"], self.rs, as_path="3 3 3",
                           next_hop=self.AS3, std_comms=[], lrg_comms=[])
        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, self.DATA["AS3_prepend2_AS2"], self.rs, as_path="3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_082_control_communities_prepend3_AS1_1_others(self):
        """{}: control communities, prepend thrice to AS1, once to others"""

        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, self.DATA["AS3_prep3AS1_1any"], self.rs, as_path="3 3 3 3",
                               next_hop=self.AS3, std_comms=[], lrg_comms=[])
        self.receive_route(self.AS2, self.DATA["AS3_prep3AS1_1any"], self.rs, as_path="3 3",
                           next_hop=self.AS3, std_comms=[], lrg_comms=[])

    def test_083_control_communities_AS3_noexport_to_any(self):
        """{}: control communities, NO_EXPORT to any"""

        for inst in (self.AS1_1, self.AS1_2, self.AS2):
            self.receive_route(inst, self.DATA["AS3_noexport_any"], self.rs,
                               as_path="3", next_hop=self.AS3,
                               std_comms=["65535:65281"],
                               lrg_comms=[], ext_comms=[])
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS101, self.DATA["AS3_noexport_any"])

    def test_083_control_communities_AS3_noexport_to_AS1(self):
        """{}: control communities, NO_EXPORT to AS1"""

        pref = self.DATA["AS3_noexport_AS1"]
        for inst in (self.AS1_1, self.AS1_2):
            self.receive_route(inst, pref, self.rs, as_path="3", next_hop=self.AS3,
                               std_comms=["65535:65281"],
                               lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS2, pref, self.rs, as_path="3 3 3 3", next_hop=self.AS3,
                           std_comms=[],
                           lrg_comms=[], ext_comms=[])
        self.receive_route(self.AS101, pref, as_path="2 3 3 3 3")
        with self.assertRaisesRegexp(AssertionError, "Routes not found."):
            self.receive_route(self.AS101, pref, as_path="1 3")

    def test_100_prefixes_received_by_clients_AS1_1(self):
        """{}: prefixes received by clients: AS1_1"""
        self.receive_route(self.AS1_1, self.DATA["AS2_good1"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_1, self.DATA["AS2_good2"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_1, self.DATA["AS2_nonclient_nexthop1"],
                           self.rs, as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop1_nh"])

    def test_100_prefixes_received_by_clients_AS1_2(self):
        """{}: prefixes received by clients: AS1_2"""

        self.receive_route(self.AS1_2, self.DATA["AS2_good1"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_2, self.DATA["AS2_good2"], self.rs, as_path="2", next_hop=self.AS2)
        self.receive_route(self.AS1_2, self.DATA["AS2_nonclient_nexthop1"],
                           self.rs, as_path="2",
                           next_hop=self.DATA["AS2_nonclient_nexthop1_nh"])

    def test_100_prefixes_received_by_clients_AS2(self):
        """{}: prefixes received by clients: AS2"""

        self.receive_route(self.AS2, self.DATA["AS1_good1"], self.rs, as_path="1", next_hop=self.AS1_1)
        self.receive_route(self.AS2, self.DATA["AS1_good2"], self.rs, as_path="1", next_hop=self.AS1_1)
        self.receive_route(self.AS2, self.DATA["AS1_good3"], self.rs, as_path="1", next_hop=self.AS1_1)

    def test_100_prefixes_received_by_clients_AS3(self):
        """{}: prefixes received by clients: AS3"""

        # AS3 has prepend_rs_as, so all the prefixes received from the rs
        # have AS_PATH "999 x"

        # prefix announced by AS1_2 only but with AS1_1 next-hop
        self.receive_route(self.AS3, self.DATA["AS1_good3"], self.rs, as_path="999 1", next_hop=self.AS1_1)

        # AS3 must receive prefixes from AS2
        self.receive_route(self.AS3, self.DATA["AS2_good1"], self.rs, as_path="999 2", next_hop=self.AS2)
        self.receive_route(self.AS3, self.DATA["AS2_good2"], self.rs, as_path="999 2", next_hop=self.AS2)

        # AS3 must receive prefixes from AS1 (this test only verifies
        # that the following prefixes reach AS3, without taking into account
        # AS_PATH nor NEXT_HOP; the "_with_ADD_PATH" version of this test
        # also verifies that these prefixes are received twice by AS3)
        for pref in ("AS101_good1", "AS101_bad_std_comm", "AS101_bad_lrg_comm",
                     "AS101_other_s_comm", "AS101_other_l_comm",
                     "AS101_bad_good_comms"):
            self.receive_route(self.AS3, self.DATA[pref], self.rs)

    def test_100_prefixes_received_by_clients_AS3_with_ADD_PATH(self):
        """{}: prefixes received by clients: AS3 (with ADD-PATH)"""

        if isinstance(self.rs, OpenBGPDInstance):
            raise unittest.SkipTest("ADD-PATH not supported by OpenBGPD")

        # AS3 has prepend_rs_as, so all the prefixes received from the rs
        # have AS_PATH "999 x"

        # AS3 must receive prefixes from both AS1_1 and AS1_2 because the
        # session on the rs is configured with ADD-PATH
        self.receive_route(self.AS3, self.DATA["AS1_good1"], self.rs, as_path="999 1", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS1_good1"], self.rs, as_path="999 1", next_hop=self.AS1_2)
        self.receive_route(self.AS3, self.DATA["AS1_good2"], self.rs, as_path="999 1", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS1_good2"], self.rs, as_path="999 1", next_hop=self.AS1_2)

        # AS101 announces its prefixes to both AS1 (AS1_1 and AS1_2 clients) and AS2.
        # AS101 prefixes received by AS1_2 are rejected by rs because of next-hop-policy failure.
        # The other prefixes should be received twice by AS3 because ADD-PATH
        # is configured for the session on the rs.
        self.receive_route(self.AS3, self.DATA["AS101_good1"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS101_good1"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route(self.AS3, self.DATA["AS101_bad_std_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS101_bad_std_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route(self.AS3, self.DATA["AS101_bad_lrg_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1)
        self.receive_route(self.AS3, self.DATA["AS101_bad_lrg_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2)

        self.receive_route(self.AS3, self.DATA["AS101_other_s_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1, std_comms=["888:0"])
        self.receive_route(self.AS3, self.DATA["AS101_other_s_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2, std_comms=["888:0"])

        self.receive_route(self.AS3, self.DATA["AS101_other_l_comm"], self.rs, as_path="999 1 101", next_hop=self.AS1_1, lrg_comms=["888:0:0"])
        self.receive_route(self.AS3, self.DATA["AS101_other_l_comm"], self.rs, as_path="999 2 101", next_hop=self.AS2, lrg_comms=["888:0:0"])

        self.receive_route(self.AS3, self.DATA["AS101_bad_good_comms"], self.rs, as_path="999 1 101", next_hop=self.AS1_1,
                           std_comms=["777:0"], lrg_comms=["777:0:0"])
        self.receive_route(self.AS3, self.DATA["AS101_bad_good_comms"], self.rs, as_path="999 2 101", next_hop=self.AS2,
                           std_comms=["777:0"], lrg_comms=["777:0:0"])

class BasicScenario_TagRejectPolicy(LiveScenario_TagRejectPolicy):

    def test_042_bad_prefixes_received_by_rs_bogon_wrong_tag(self):
        """{}: bad prefixes received by rs: bogon (wrong tag)"""
        with self.assertRaisesRegexp(AssertionError, "real reasons 2, expected reason 1."):
            self.receive_route(self.rs, self.DATA["bogon1"], self.AS1_1,
                               as_path="1", next_hop=self.AS1_1,
                               filtered=True, reject_reason=1)

    def test_042_bad_prefixes_received_by_rs_global_blacklist_wrong_tag(self):
        """{}: bad prefixes received by rs: global blacklist (wrong tag)"""
        with self.assertRaisesRegexp(AssertionError, "real reasons 3, expected reason 1."):
            self.receive_route(self.rs, self.DATA["local1"], self.AS1_1,
                               as_path="1", next_hop=self.AS1_1,
                               filtered=True, reject_reason=1)

class BasicScenarioBIRD(BasicScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                        cfg_roas="roas{}.yml".format(cls.IP_VER),
                                        local_files=["footer{}".format(cls.IP_VER)]),
                    "/etc/bird/bird.conf"
                ),
                (
                    cls.use_static_file("bird_local_file.local{}".format(cls.IP_VER)),
                    "/etc/bird/footer{}.local".format(cls.IP_VER)
                )
            ],
        )

class BasicScenarioOpenBGPD(BasicScenario_TagRejectPolicy, BasicScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder
    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     local_files_dir="/etc/bgpd",
                                     local_files=["post-clients", "post-filters"],
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                ),
                (
                    cls.use_static_file("openbgpd_post-clients.local"),
                    "/etc/bgpd/post-clients.local"
                ),
                (
                    cls.use_static_file("openbgpd_post-filters.local"),
                    "/etc/bgpd/post-filters.local"
                )
            ]
        )

class BasicScenarioOpenBGPD60(BasicScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = "6.0"

class BasicScenarioOpenBGPD61(BasicScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = "6.1"



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest

from pierky.arouteserver.peering_db import PeeringDBNet
from pierky.arouteserver.errors import PeeringDBNoInfoError 
from pierky.arouteserver.tests.mock_peeringdb import mock_peering_db


class TestPeeringDBInfo(unittest.TestCase):

    def setUp(self):

        mock_peering_db(os.path.dirname(__file__) + "/peeringdb_data")

    def test_net1(self):
        """PeeringDB network: get data"""
        net = PeeringDBNet(1)
        self.assertEqual(net.info_prefixes4, 20)
        self.assertEqual(net.info_prefixes6, 10)

    def test_no_data(self):
        """PeeringDB network: missing data"""
        with self.assertRaises(PeeringDBNoInfoError):
            net = PeeringDBNet(2)



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import ARouteServerCommand

from ..config.program import program_config

class SetupTemplatesCommand(ARouteServerCommand):

    COMMAND_NAME = "setup-templates"
    COMMAND_HELP = ("Install template files distributed "
                    "within the package. Useful after an "
                    "upgrade to sync the local templates "
                    "to those used by the new version.")
    NEEDS_CONFIG = True

    def run(self):
        return program_config.setup_templates()



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .cfg_base import TestConfigParserBase 
from pierky.arouteserver.config.roa import ConfigParserROAEntries


class TestConfigParserROA(TestConfigParserBase):

    FILE_PATH = "tests/static/test_cfg_roa_entries.yml"
    CONFIG_PARSER_CLASS = ConfigParserROAEntries
    SHORT_DESCR = "ROAs entries config parser"

    def test_valid_cfg(self):
        """{}: valid configuration"""
        self._contains_err()

    def test_unknown_statement(self):
        """{}: unknown statement"""
        self.cfg[0]["test"] = 1
        self._contains_err("Error in ROAs definition: Unknown statement 'test' in ROA entry definition.")

    def test_valid_roas(self):
        """{}: valid ROA entries"""
        roa = self.cfg[0]
        self.assertEqual(roa["asn"], 1)
        self.assertEqual(roa["prefix"]["prefix"], "10.0.1.0")
        self.assertEqual(roa["prefix"]["length"], 24)
        self.assertEqual(roa["prefix"]["exact"], True)

        roa = self.cfg[1]
        self.assertEqual(roa["asn"], 2)
        self.assertEqual(roa["prefix"]["prefix"], "10.1.0.0")
        self.assertEqual(roa["prefix"]["length"], 16)
        self.assertEqual(roa["prefix"]["le"], 24)
        self.assertEqual(roa["prefix"]["exact"], False)

        roa = self.cfg[2]
        self.assertEqual(roa["asn"], 3)
        self.assertEqual(roa["prefix"]["prefix"], "10.2.0.0")
        self.assertEqual(roa["prefix"]["length"], 24)
        self.assertEqual(roa["prefix"]["exact"], True)

    def test_invalid_ge(self):
        """{}: invalid ge"""
	cfg = [
            "roas:",
	    "- prefix:",
            "    prefix: 10.0.1.0",
            "    length: 24",
            "    ge: 25",
            "  asn: 1"
        ]
        self.cfg._load_from_yaml("\n".join(cfg))
        self._contains_err("Error in ROAs definition: ROA prefix 'ge' must be equal to the prefix length: 25 != 24 for prefix 10.0.1.0.")



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from base import RichConfigExampleScenarioBIRD
from data4 import RichConfigExampleScenario_Data4
from pierky.arouteserver.tests.live_tests.bird import BIRDInstanceIPv4

class RichConfigExampleScenario_BIRDIPv4(RichConfigExampleScenario_Data4,
                                         RichConfigExampleScenarioBIRD):
    __test__ = True

    SHORT_DESCR = "Live test, BIRD, examples, rich config, IPv4"
    RS_INSTANCE_CLASS = BIRDInstanceIPv4
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv4
    IP_VER = 4



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import BaseConfigEnricher, BaseConfigEnricherThread
from ..errors import BuilderError, ARouteServerError, \
                    PeeringDBError, PeeringDBNoInfoError
from ..peering_db import PeeringDBNet

class PeeringDBConfigEnricher_WorkerThread(BaseConfigEnricherThread):

    DESCR = "PeeringDB"

    def __init__(self, *args, **kwargs):
        BaseConfigEnricherThread.__init__(self, *args, **kwargs)

        self.ip_ver = None
        self.cfg_general = None
        self.cache_dir = None
        self.cache_expiry = None

    def do_task(self, task):
        client = task

        client_max_prefix = client["cfg"]["filtering"]["max_prefix"]
        if not client_max_prefix["action"]:
            # No max-prefix action given for this client:
            # no needs to know its max-pref limit.
            return None

        res = {"limit4": None, "limit6": None}

        for ip_ver in (4, 6):
            if self.ip_ver is not None and self.ip_ver != ip_ver:
                # The current AF is not used.
                continue

            if client_max_prefix["limit_ipv{}".format(ip_ver)]:
                # Client uses a specific limit:
                # no needs to gather info from PeeringDB for
                # the current address family.
                res["limit{}".format(ip_ver)] = \
                    client_max_prefix["limit_ipv{}".format(ip_ver)]
                continue

            general_limit = self.cfg_general["filtering"]["max_prefix"]["general_limit_ipv{}".format(ip_ver)]

            if not client_max_prefix["peering_db"]:
                # PeeringDB disabled for this client:
                # using general limit.
                res["limit{}".format(ip_ver)] = general_limit
                continue

            try:
                peeringdb_limit = None
                net = PeeringDBNet(client["asn"],
                                   cache_dir=self.cache_dir,
                                   cache_expiry=self.cache_expiry)
                if ip_ver == 4:
                    peeringdb_limit = net.info_prefixes4
                else:
                    peeringdb_limit = net.info_prefixes6

                res["limit{}".format(ip_ver)] = peeringdb_limit or general_limit

            except PeeringDBNoInfoError:
                # No data found on PeeringDB.
                logging.debug("No data found on PeeringDB "
                              "for AS{} while looking for "
                              "max-prefix limit.".format(client["asn"]))
                pass
            except PeeringDBError as e:
                raise BuilderError(
                    "An error occurred while retrieving info from PeeringDB "
                    "for ASN {}: {}".format(
                        client["asn"], str(e) or "error unknown"
                    )
                )

        return res["limit4"], res["limit6"]

    def save_data(self, task, data):
        client = task
        limit4, limit6 = data
        if limit4:
            client["cfg"]["filtering"]["max_prefix"]["limit_ipv4"] = limit4
        if limit6:
            client["cfg"]["filtering"]["max_prefix"]["limit_ipv6"] = limit6

class PeeringDBConfigEnricher(BaseConfigEnricher):

    WORKER_THREAD_CLASS = PeeringDBConfigEnricher_WorkerThread

    def _config_thread(self, thread):
        thread.ip_ver = self.builder.ip_ver
        thread.cfg_general = self.builder.cfg_general
        thread.cache_dir = self.builder.cache_dir
        thread.cache_expiry = self.builder.cache_expiry

    def add_tasks(self):
        # Enqueuing tasks.
        for client in self.builder.cfg_clients.cfg["clients"]:
            self.tasks_q.put(client)



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import mock

from ..cached_objects import CachedObject
from ..peering_db import PeeringDBInfo, PeeringDBNet


def mock_peering_db(data_dir=None):

    def get_data_from_peeringdb(self):
        path = "{}/{}".format(
            data_dir or os.path.dirname(__file__),
            self._get_peeringdb_url()
        )
        with open(path, "r") as f:
            return json.load(f)

    mock_get_data_from_peeringdb = mock.patch.object(
        PeeringDBInfo, "_get_data_from_peeringdb", autospec=True
    ).start()
    mock_get_data_from_peeringdb.side_effect = get_data_from_peeringdb

    def get_url_net(self):
        return "net_{}.json".format(self.asn)

    mock_get_url_net = mock.patch.object(
        PeeringDBNet, "_get_peeringdb_url", autospec=True
    ).start()
    mock_get_url_net.side_effect = get_url_net

    def load_data_from_cache(self):
        return False

    mock_load_data_from_cache = mock.patch.object(
        CachedObject, "load_data_from_cache", autospec=True
    ).start()
    mock_load_data_from_cache.side_effect = load_data_from_cache

    def save_data_to_cache(self):
        return

    mock_save_data_to_cache = mock.patch.object(
        CachedObject, "save_data_to_cache", autospec=True
    ).start()
    mock_save_data_to_cache.side_effect = save_data_to_cache



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re

from base import ARouteServerCommand
from ..ask import ask, ask_yes_no
from ..config.program import program_config
from ..errors import ARouteServerError
from ..resources import get_live_test_skeleton_dir

class InitScenarioCommand(ARouteServerCommand):

    COMMAND_NAME = "init-scenario"
    COMMAND_HELP = ("Initialize a new live test scenario by copying "
                    "files from the skeleton example.")
    NEEDS_CONFIG = True

    @classmethod
    def add_arguments(cls, parser):
        super(InitScenarioCommand, cls).add_arguments(parser)

        parser.add_argument(
            "dest_dir",
            help="Destination directory for the new scenario")

    def run(self):
        print("Please provide an identifier for the new scenario, "
              "something that will be used as the name of the "
              "Python classes where the test functions will be "
              "implemented.")
        print("Examples: TagASSet, BGPCommunities, MaxPrefix, PathHiding")
        print("")
        res, class_name = ask("Scenario class name:")

        if not res:
            print("")
            print("Aborted.")
            return False

        if class_name.lower().endswith("scenario"):
            class_name = class_name[:-8]

        if not re.match("[a-z_][a-z_0-9 ]+", class_name, re.IGNORECASE):
            print("")
            print("Invalid name. "
                  "It must be in the format "
                  "(letter|\"_\") (letter | digit | \"_\")*")
            print("Aborted.")
            return False

        class_name = "{}Scenario".format(class_name.title()).replace(" ", "")

        skeleton_dir = get_live_test_skeleton_dir()
        dest_dir = os.path.expanduser(self.args.dest_dir)
        templates_dir = program_config.get("templates_dir")

        if os.path.exists(dest_dir):
            raise ARouteServerError(
                "The directory {} already exists".format(dest_dir)
            )

        res, yes_or_no = ask_yes_no(
            "The {} directory will be created: proceed?".format(dest_dir),
            default="yes"
        )

        print("")

        if not res or yes_or_no != "yes":
            print("Aborted.")
            return False

        try:
            os.makedirs(dest_dir)
        except Exception as e:
            raise ARouteServerError(
                "An error occurred while creating {}: {}".format(
                    dest_dir, str(e)
                )
            )

        for speaker in ("bird", "openbgpd"):
            link_src = "{}/{}".format(templates_dir, speaker)
            link_dst = "{}/{}".format(dest_dir, speaker)
            try:
                os.symlink(link_src, link_dst)
            except Exception as e:
                raise ARouteServerError(
                    "An error occurred while creating a link "
                    "to the '{}' templates directory: link "
                    "source {}, link destination {}: {}".format(
                        speaker, link_src, link_dst, str(e)
                    )
                )
            print("A link to the '{}' templates directory {} "
                  "has been created in {}.".format(
                    speaker, link_src, link_dst
                )
            )

        for filename in os.listdir(skeleton_dir):
            src_path = "{}/{}".format(skeleton_dir, filename)
            dst_path = "{}/{}".format(dest_dir, filename)
            if not os.path.isfile(src_path):
                continue
            try:
                with open(src_path, "r") as s:
                    with open(dst_path, "w") as d:
                        content = s.read()
                        if filename in ("base.py", "test_bird4.py",
                                        "test_bird6.py", "test_openbgpd4.py",
                                        "test_openbgpd6.py"):
                            content = content.replace(
                                "SkeletonScenario", class_name
                            )
                            content = content.replace(
                                ", skeleton,", ", {},".format(class_name)
                            )
                        d.write(content)
            except Exception as e:
                raise ARouteServerError(
                    "An error occurred while copying {} "
                    "into {}: {}".format(
                        src_path, dst_path, str(e)
                    )
                )

        print("A new live test scenario has been initialized "
              "in the {} directory.\n\nFor details: {}".format(
                  dest_dir,
                  "https://arouteserver.readthedocs.io/en/latest/LIVETESTS.html"
                  "#how-to-build-custom-scenarios"
              )
        )



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from base import TagASSetScenario_WithAS_SETs, TagASSetScenario_EmptyAS_SETs, \
                 TagASSetScenarioBIRD
from data6 import TagASSetScenario_Data6
from pierky.arouteserver.tests.live_tests.bird import BIRDInstanceIPv6

class TagASSetScenario_WithAS_SETs_BIRDIPv6(TagASSetScenario_WithAS_SETs,
                                            TagASSetScenario_Data6,
                                            TagASSetScenarioBIRD):
    __test__ = True

    RS_INSTANCE_CLASS = BIRDInstanceIPv6
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv6
    IP_VER = 6

    SHORT_DESCR = "Live test, BIRD, tag prefix/origin in AS-SET, IPv6"

class TagASSetScenario_EmptyAS_SETs_BIRDIPv6(TagASSetScenario_EmptyAS_SETs,
                                             TagASSetScenario_Data6,
                                             TagASSetScenarioBIRD):
    __test__ = True

    RS_INSTANCE_CLASS = BIRDInstanceIPv6
    CLIENT_INSTANCE_CLASS = BIRDInstanceIPv6
    IP_VER = 6

    SHORT_DESCR = "Live test, BIRD, tag prefix/origin empty AS-SET, IPv6"



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import time

from .errors import CachedObjectsError


class CachedObject(object):

    DEFAULT_EXPIRY = 43200

    def __init__(self, **kwargs):
        self.cache_dir = kwargs.get("cache_dir", "var")
        if not self.cache_dir:
            raise CachedObjectsError("Missing cache directory")

        self.cache_expiry_time = kwargs.get("cache_expiry",
                                            self.DEFAULT_EXPIRY)
        self.raw_data = None

    def _get_object_filename(self):
        raise NotImplementedError()

    def _get_object_filepath(self):
        return os.path.join(self.cache_dir, self._get_object_filename())

    def load_data_from_cache(self):
        file_path = self._get_object_filepath()

        if not os.path.isfile(file_path):
            return False

        try:
            with open(file_path, "r") as f:
                data = json.load(f)
        except:
            logging.error(
                "Error while reading data from cache: {}".format(file_path)
            )
            return False

        if not "ts" in data:
            return False
        if not "data" in data:
            return False

        epoch_time = int(time.time())

        if data["ts"] <= epoch_time - self.cache_expiry_time:
            return False

        self.raw_data = data["data"]
        return True

    def _get_data(self):
        raise NotImplementedError()

    def load_data(self):
        if self.load_data_from_cache():
            logging.debug("Cache hit: {}".format(self._get_object_filepath()))
            return

        self.raw_data = self._get_data()

        self.save_data_to_cache()

    def save_data_to_cache(self):
        file_path = self._get_object_filepath()

        epoch_time = int(time.time())

        cache_data = {
            "ts": epoch_time,
            "data": self.raw_data
        }

        if self.raw_data:
            try:
                if not os.path.exists(os.path.dirname(file_path)):
                    os.makedirs(os.path.dirname(file_path))
                with open(file_path, "w") as f:
                    json.dump(cache_data, f)
            except Exception as e:
                raise CachedObjectsError(
                    "Error while saving data to the cache: {}".format(str(e))
                )



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  Ifnot, see <http://www.gnu.org/licenses/>.

import os
import pkg_resources

from errors import ResourceNotFoundError


def get_local_dir(dirname):
    local_path = dirname
    if os.path.isdir(local_path):
        return local_path

    pkg_path = pkg_resources.resource_filename("pierky.arouteserver", dirname)
    if os.path.isdir(pkg_path):
        return pkg_path

    raise ResourceNotFoundError(
        "Can't find '{}' directory neither at {} nor at {}".format(
            dirname, local_path, pkg_path
        )
    )

def get_config_dir():
    return get_local_dir("config.d")

def get_templates_dir():
    return get_local_dir("templates")

def get_live_test_skeleton_dir():
    return get_local_dir("tests/live_tests/skeleton")

def get_config_file(filename):
    config_dir = get_config_dir()
    path = os.path.join(config_dir, filename)
    if os.path.isfile(path):
        return path

    raise ResourceNotFoundError(
        "Can't find '{}' configuration file at {}".format(
            filename, config_dir
        )
    )



# Copyright (C) 2017 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ipaddr
import json
import os
import logging
import subprocess
import time

from .cached_objects import CachedObject
from .config.validators import ValidatorPrefixListEntry
from .errors import IRRDBToolsError


class IRRDBTools(CachedObject):

    BGPQ3_DEFAULT_HOST = "rr.ntt.net"
    BGPQ3_DEFAULT_SOURCES = ("RIPE,APNIC,AFRINIC,ARIN,NTTCOM,ALTDB,"
                             "BBOI,BELL,GT,JPIRR,LEVEL3,RADB,RGNET,"
                             "SAVVIS,TC")

    def __init__(self, *args, **kwargs):
        CachedObject.__init__(self, *args, **kwargs)
        self.bgpq3_path = kwargs.get("bgpq3_path")
        self.bgpq3_host = kwargs.get("bgpq3_host", self.BGPQ3_DEFAULT_HOST)
        self.bgpq3_sources = kwargs.get("bgpq3_sources",
                                        self.BGPQ3_DEFAULT_SOURCES)

class ASSet(IRRDBTools):

    def __init__(self, object_name, **kwargs):
        IRRDBTools.__init__(self, **kwargs)
        self.object_name = object_name

        logging.debug("Getting origin ASNs for "
                      "{} from IRRdb".format(self.object_name))

        self.load_data()

        # list of int
        self.asns = self.raw_data

    def _get_object_filename(self):
        return "{}-as_set.json".format(self.object_name)

    def _get_data(self):
        cmd = [self.bgpq3_path]
        cmd += ["-h", self.bgpq3_host]
        cmd += ["-S", self.bgpq3_sources]
        cmd += ["-3"]
        cmd += ["-j"]
        cmd += ["-f", "1"]
        cmd += ["-l", "asn_list"]
        cmd += [self.object_name]

        try:
            out = subprocess.check_output(cmd)
        except Exception as e:
            raise IRRDBToolsError(
                "Can't get list of authorized ASNs for {}: {}".format(
                    self.object_name, str(e)
                )
            )

        try:
            data = json.loads(out)
        except Exception as e:
            raise IRRDBToolsError(
                "Error while parsing bgpq3 output "
                "for the following command: '{}': {}".format(
                    " ".join(cmd), str(e)
                )
            )

        return data["asn_list"]

class RSet(IRRDBTools):

    def __init__(self, object_name, ip_ver, allow_longer_prefixes, **kwargs):
        IRRDBTools.__init__(self, **kwargs)
        self.object_name = object_name
        assert ip_ver in (4, 6)
        self.ip_ver = ip_ver
        self.allow_longer_prefixes = allow_longer_prefixes

        logging.debug("Getting prefixes for {} IPv{} "
                      "from IRRdb".format(self.object_name, self.ip_ver))

        self.load_data()

        # list of dict as returned by ValidatorPrefixListEntry
        self.prefixes = self.raw_data

    def _get_object_filename(self):
        return "{}-r_set-ipv{}.json".format(self.object_name, self.ip_ver)

    def _get_data(self):
        cmd = [self.bgpq3_path]
        cmd += ["-h", self.bgpq3_host]
        cmd += ["-S", self.bgpq3_sources]
        cmd += ["-3"]
        cmd += ["-4"] if self.ip_ver == 4 else ["-6"]
        cmd += ["-A"]
        cmd += ["-j"]
        cmd += ["-l", "prefix_list"]
        if self.allow_longer_prefixes:
            cmd += ["-R"]
            cmd += ["32"] if self.ip_ver == 4 else ["128"]
        cmd += [self.object_name]

        try:
            out = subprocess.check_output(cmd)
        except Exception as e:
            raise IRRDBToolsError(
                "Can't get authorized prefix list for {} IPv{}: {}".format(
                    self.object_name, self.ip_ver, str(e)
                )
            )

        try:
            data = json.loads(out)
        except Exception as e:
            raise IRRDBToolsError(
                "Error while parsing bgpq3 output "
                "for the following command: '{}': {}".format(
                    " ".join(cmd), str(e)
                )
            )

        return [self._parse_prefix(prefix) for prefix in data["prefix_list"]]

    def _parse_prefix(self, raw):
        prefix = ipaddr.IPNetwork(raw["prefix"])
        res = {
            "prefix": str(prefix.ip),
            "length": prefix.prefixlen,
            "exact": raw["exact"] if "exact" in raw else False,
            "comment": self.object_name
        }
        if res["exact"]:
            res["ge"] = None
            res["le"] = None
        else:
            if "greater-equal" in raw:
                res["ge"] = raw["greater-equal"]
            else:
                res["ge"] = None

            if "less-equal" in raw:
                res["le"] = raw["less-equal"]
            else:
                res["le"] = None

        return ValidatorPrefixListEntry().validate(res)
